"""Dump strongly-typed containers into loosely-typed objects."""

from dataclasses import asdict, fields, is_dataclass
from typing import Any, Mapping, Sequence

from .overrides import get_override
from .typedefs import MISSING, NamedTupleType


def _filter_keep(value: Any, keep_always: bool) -> bool:
    """Determine whether we should keep a value in serialized output."""
    if keep_always:
        return True
    return value is not MISSING


def _identity(value: Any) -> Any:
    """Simple identity function, returning original value."""
    return value


def dump(obj: Any, convert_missing_to_none: bool = False) -> Any:
    """Serialize the object into a less-typed form.

    If serializing a dataclass and `transform_dump` metadata exists in a
    `dataclasses.field`, its value is assumed to be function whose result is
    serialized before being passed recursively down the chain.

    :param convert_missing_to_none: retain missing property's key and convert
        its value to None. This will keep all keys when serializing, useful if
        you want to keep all column names and assign value of `None` to missing
        columns.

    Serialize from: - `x` -> `y`:
      - `dataclass` -> `Dict`
      - `NamedTuple` -> `Dict`
      - `str` -> `str`
      - `Sequence` -> `List`
      - `Mapping` -> `Dicts`
      * if `convert_missing_to_none` is True:
        - `MISSING` -> `None`
      * else
        - `MISSING` keys are filtered out and missing values are kept as-is
      - `Anything else` -> `itself`
    """
    # pylint: disable=too-many-return-statements
    if is_dataclass(obj):
        custom_dump = {
            f.name: get_override(f.metadata.get("serdelicacy"))
            for f in fields(obj)
        }
        return {
            key: dump(__value_converted, convert_missing_to_none)
            for key, value in asdict(obj).items()
            if _filter_keep(
                (__value_converted := custom_dump[key].transform_dump(value)),
                convert_missing_to_none,
            )
        }
    if isinstance(obj, NamedTupleType):
        return {
            key: dump(value, convert_missing_to_none)
            for key, value in obj._asdict().items()
            if _filter_keep(value, convert_missing_to_none)
        }
    if isinstance(obj, str):
        return obj
    if isinstance(obj, Sequence):
        return [
            dump(value, convert_missing_to_none)
            for value in obj
            if _filter_keep(value, convert_missing_to_none)
        ]
    if isinstance(obj, Mapping):
        return {
            dump(key, convert_missing_to_none): dump(
                value, convert_missing_to_none
            )
            for key, value in obj.items()
            if _filter_keep(value, convert_missing_to_none)
        }
    if convert_missing_to_none and (obj is MISSING):
        return None
    return obj



"""Custom Exceptions for serdelicacy."""

import json
from typing import Any, List, NamedTuple, Type

from .typedefs import MISSING


class DepthContainer(NamedTuple):
    """Contain information."""

    constructor: Type
    key: Any
    value: Any


class SerdeError(Exception):
    """Root error for serdelicacy."""


def _tc_enc(value: str, code: int = 32) -> str:
    """Encode error message values with terminal colors.

    See:
        - <https://i.stack.imgur.com/9UVnC.png>
        - <https://stackoverflow.com/a/61273717>

    For Python's somewhat-poorly-documented encodings:
        <https://docs.python.org/3/library/codecs.html#text-encodings>

    Some color options:
        - 30: black
        - 31: red
        - 32: green
        - 33: yellow
        - 34: blue
        - 35: magenta
        - 36: cyan
        - 37: white
    """
    return f"\u001b[{code}m{value}\u001b[0m"


_K_KEY = _tc_enc("key", 32)
_K_VALUE = _tc_enc("value", 32)
_K_TYPE = _tc_enc("type", 34)
_K_INPUT = _tc_enc("input", 34)
_K_ERROR = _tc_enc("error", 31)


class DeserializeError(SerdeError):
    """Deserialization failure.

    Deserializing arbitrarily-nested JSON often results in opaque
    deserialization errors. This Exception class provides a clear,
    consistent debugging message.
    """

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        type_expected: Type,
        value_received: Any,
        depth: List[DepthContainer],
        key: Any,
        message_prefix: str = "",
        message_postfix: str = "",
        message_override: str = "",
    ):
        depth_messages = []
        for depth_item in depth:
            value = {
                _K_KEY: repr(depth_item.key),
                _K_VALUE: {
                    _K_TYPE: repr(depth_item.constructor),
                    _K_INPUT: repr(depth_item.value),
                },
            }
            if depth_item.key is MISSING:
                del value[_K_KEY]
            depth_messages.append(value)

        if message_override:
            message = message_override
        elif value_received is MISSING and key is not MISSING:
            message = f"missing required key {repr(key)}"
            depth_messages.pop()
        else:
            message = (
                message_prefix
                + f"expected {repr(type_expected)} "
                + f"but received {repr(type(value_received))} "
                + message_postfix
            )
        depth_messages[-1][_K_ERROR] = message.strip()
        depth_str = json.dumps(depth_messages, indent=2)
        super().__init__(
            f"{message}\n{depth_str}".encode().decode("unicode-escape")
        )



"""Serdelicacy overrides for dataclasses."""

from dataclasses import dataclass, field
from typing import Any, Callable, NoReturn, TypeVar, Union

T = TypeVar("T")  # pylint: disable=invalid-name


def _noreturn(value: Any) -> NoReturn:  # pylint: disable=unused-argument
    """Simple 1 argument function that returns nothing."""


def _id(value: T) -> T:
    """Identity function."""
    return value


@dataclass(frozen=True)
class Override:
    """Metadata user-defined overrides for dataclass serde.

    :attr dataclass_validate: argument provided only by upstream dataclasses.
        An internal implementation detail, but interesting nonetheless.
        Function either returns `True` on positive validation / `False` on
        non-validation, or returns nothing at all and instead relies on the
        raising of exceptions to indicate whether validation passed for failed.
    :attr dataclass_transform_load: when deserializing, evaluated on an object
        before the object is recursively examined.
    :attr dataclass_transform_postload: when deserializing, evaluated on an
        object after the object has been recursively examined. When possible,
        the `transform_load` should be preferred over `transform_postload`, but
        there are situations where `transform_postload` is useful.
    :attr dataclass_transform_dump: when serializing a dataclass, called on
        value before it is recursively serialized
    """

    validate: Union[
        Callable[[Any], NoReturn],
        Callable[[Any], bool],
        Callable[[Any], None],
    ] = field(default=_noreturn)
    transform_load: Callable[[Any], Any] = field(default=_id)
    transform_postload: Callable[[Any], Any] = field(default=_id)
    transform_dump: Callable[[Any], Any] = field(default=_id)


DEFAULT_OVERRIDE = Override()


def get_override(value: Any) -> Override:
    """Perform validation for override value."""
    if value is None:
        return DEFAULT_OVERRIDE
    if isinstance(value, Override):
        return value
    raise TypeError(
        "dataclasses field error for metadata key 'serdelicacy': value "
        f"{repr(value)} is not the correct type. Expected {repr(Override)}"
    )



"""serdelicacy: load and dump data into strongly-typed data structures

This library has two goals:

1. "Deserialize" unstructured Python types into structured, type-hinted Python
    types (dataclasses.dataclass, typing.NamedTuples).
2. "Serialize" structured, type-hinted Python objects into unstructured Python
    types (eg, the reverse).
"""

from .deserialize import load
from .errors import DeserializeError
from .overrides import Override
from .serialize import dump
from .typedefs import OptionalProperty, get, is_missing



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['serdelicacy']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'serdelicacy',
    'version': '0.16.0',
    'description': 'Serialize/deserialize Python objects from/to typed structures.',
    'long_description': '# serdelicacy\n\n[![image-version](https://img.shields.io/pypi/v/serdelicacy.svg)](https://python.org/pypi/serdelicacy)\n[![image-license](https://img.shields.io/pypi/l/serdelicacy.svg)](https://python.org/pypi/serdelicacy)\n[![image](https://img.shields.io/pypi/pyversions/serdelicacy.svg)](https://python.org/pypi/serdelicacy)\n[![image-ci](https://github.com/pappasam/serdelicacy/workflows/serdelicacy%20ci/badge.svg)](https://github.com/pappasam/serdelicacy/actions?query=workflow%3A%22serdelicacy+ci%22)\n\nSerialize (`serdelicacy.dump`) and deserialize (`serdelicacy.load`) from/to strongly-typed, native Python data structures.\n\n## Features\n\n1. Effortless deserialization of unstructured Python types into structured, type-hinted Python types (`dataclasses.dataclass`, `typing.NamedTuple`)\n2. Effortless serialization of structured, type-hinted Python objects into unstructured Python types (eg, the reverse)\n3. Clear error messages when serde fails at runtime\n4. No inherited, non-standard types. dataclasses, NamedTuples, and other standard Python types are bread and butter\n5. Editor support: I like my autocompletion, so I jump through lots of hoops to make this library compatible with Jedi\n6. Handle [optional properties](https://www.typescriptlang.org/docs/handbook/interfaces.html#optional-properties) with a domain-specific `serdelicacy.OptionalProperty`\n7. Enable customization through sophisticated validation, deserialization overrides, and serialization overrides for dataclasses.\n8. Require no 3rd party dependencies; Python 3.8+\n\n## Installation\n\n```bash\n# With pip\npip install serdelicacy\n\n# With poetry\npoetry add serdelicacy\n```\n\n## Usage\n\nSee [examples folder](https://github.com/pappasam/serdelicacy/tree/master/example).\n\n## Validation / transformation for dataclasses\n\nCustomization override options are available for validations and transformations on both deserialization and serialization. Custom overrides are available for `dataclasses` through the `metadata` argument to the `dataclasses.field` function:\n\n```python\nfrom dataclasses import dataclass, field\n\nimport serdelicacy\nfrom serdelicacy import Override\n\ndef _is_long_enough(value) -> None:\n    if len(value) < 4:\n        raise ValueError(f"\'{value}\' is not enough characters")\n\nVALUE = {"firstname": "richard", "lastname": "spencerson"}\n\n@dataclass\nclass Person:\n    firstname: str = field(\n        metadata={\n            "serdelicacy": Override(\n                validate=_is_long_enough,\n                transform_load=str.title,\n            )\n        }\n    )\n    lastname: str = field(\n        metadata={\n            "serdelicacy": Override(\n                validate=_is_long_enough,\n                transform_load=str.title,\n                transform_dump=str.upper,\n            )\n        }\n    )\n\nprint(serdelicacy.load(VALUE, Person))\n```\n\nAs suggested by the Python [dataclasses.field documentation](https://docs.python.org/3/library/dataclasses.html#dataclasses.field), all `serdelicacy`-related field metadata is namespaced to 1 dictionary key: `serdelicacy`. Its value should be of type `serdelicacy.Override`, a dataclass itself whose fields are the following:\n\n- `validate`: `Callable[[Any], NoReturn], Callable[[Any], bool]`: a function that either a) returns a boolean where False indicates failed validation or b) nothing, but raises Python exceptions on validation failure. Is executed as the final step of a value\'s load, after all transformations have been completed. By default, this is a function that does nothing.\n- `transform_load`: `Callable[[Any], Any]`. This transformation is executed before any other loading takes place. By default, this is an [identity function](https://en.wikipedia.org/wiki/Identity_function)\n- `transform_postload`: this should be `Callable[[T], T]]`, where `T` is the type of the field. This transformation is executed after all recursive loading takes place as the final step before the value is returned for upstream processing. By default, this is an [identity function](https://en.wikipedia.org/wiki/Identity_function)\n\nFinally, you may not need to use these tools initially, but if you have strict validation or transformation requirements on your project, you\'ll be extremely happy they\'re here\n\n## FAQ\n\n### My JSON keys contain whitespace, etc\n\nSimple solution: use `typeing.TypeDict`\'s [backwards-compatibility syntax](https://www.python.org/dev/peps/pep-0589/#alternative-syntax).\n\n```python\nfrom pprint import pprint\nfrom typing import List, TypedDict\n\nimport serdelicacy\nfrom serdelicacy import OptionalProperty\n\nDATA = [\n    {\n        "weird, key": 1,\n        "normal": 2,\n    },\n    {\n        "normal": 3,\n    },\n]\n\nDataItem = TypedDict(\n    "DataItem",\n    {\n        "weird, key": OptionalProperty[int],\n        "normal": int,\n    },\n)\n\nLOADED = serdelicacy.load(DATA, List[DataItem])\n\nprint("Loaded data:")\npprint(LOADED)\n\nprint("Re-serialized data:")\npprint(serdelicacy.dump(LOADED))\n```\n\nThis prints the following to the console.\n\n```text\nLoaded data:\n[{\'normal\': 2, \'weird, key\': 1},\n {\'normal\': 3, \'weird, key\': <Missing property>}]\nRe-serialized data:\n[{\'normal\': 2, \'weird, key\': 1}, {\'normal\': 3}]\n```\n\nTry changing values in your JSON data; you\'ll get runtime errors if your data does not conform to the above schema. Additionally, `mypy` should call out any misused variable keys / types. In short, this has enabled a type-safe load and a perfectly sane dump.\n\n## Local Development\n\nLocal development for this project is simple.\n\n**Dependencies**\n\nInstall the following tools manually.\n\n- [Poetry](https://github.com/sdispater/poetry#installation)\n- [GNU Make](https://www.gnu.org/software/make/)\n\n_Recommended_\n\n- [asdf](https://github.com/asdf-vm/asdf)\n\n**Set up development environment**\n\n```bash\nmake setup\n```\n\n**Run Tests**\n\n```bash\nmake test\n```\n\n## Notes\n\n- Initially inspired by [undictify](https://github.com/Dobiasd/undictify) and a PR I helped with. serdelicacy\'s goals are different; it\'s focused on serde instead of general function signature overrides.\n- I also notice some striking similarities with a library called [typedload](https://github.com/ltworf/typedload) (great minds think alike, I guess :p). I renamed my top-level functions to "load" and "dump" in typedload\'s homage. Unfortunately, as of version `1.20`, typedload does not handle all types of dataclasses elegantly (mainly, InitVar). Since typedload supports Python 3.5+, it never will elegantly handle all dataclasses without lots of unfortunate conditionals in the codebase. If you must use Python 3.7-, I suggest looking into typedload.\n\n## Written by\n\nSamuel Roeca *samuel.roeca@gmail.com*\n',
    'author': 'Sam Roeca',
    'author_email': 'samuel.roeca@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pappasam/serdelicacy',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Type definitions."""

import enum
from typing import Any, Protocol, TypeVar, Union, runtime_checkable

# pylint: disable=too-few-public-methods
# pylint: disable=missing-function-docstring

T = TypeVar("T")  # pylint: disable=invalid-name


class Missing(enum.Enum):
    """Handles missing JSON values.

    TypeScript / JavaScript /  have 2 words that indicate something is
    conceptually missing:

    - `undefined` == an attribute has not been defined. This concept is
        represented by this class
    - `null` == an attribute's value is `null`. This concept is represented by
        the Python value `None`.
    """

    token = 0

    def __repr__(self) -> str:
        return "<Missing property>"

    def __bool__(self) -> bool:
        return False


OptionalProperty = Union[Missing, T]

MISSING = Missing.token


def is_missing(value: Any) -> bool:
    """Check whether a value is `MISSING`

    Useful because it prevents users from needing to import MISSING
    constant
    """
    return value is MISSING


def get(value: Union[Missing, T], default: T) -> T:
    """Return value unless it's MISSING, in which case return default.

    Similar to `dict.get`, but operates on `OptionalProperty`, provides
    no default for default, and is typesafe.
    """
    return default if value is MISSING else value  # type: ignore


class NoResult(enum.Enum):
    """Special case type to indicate that a function returns no result.

    Useful because `None`, which most functions return implicitly, is
    `null` in JSON and therefore important for serde.
    """

    token = 0

    def __bool__(self) -> bool:
        return False


PossibleResult = Union[NoResult, T]

NO_RESULT = NoResult.token


@runtime_checkable
class NamedTupleType(Protocol):
    """NamedTuple protocol; methods only associated with namedtuples."""

    def _make(self):
        ...

    def _asdict(self):
        ...

    def _replace(self):
        ...

    def _fields(self):
        ...

    def _field_defaults(self):
        ...
